/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.awt.Point;
import java.awt.Dimension;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Random;
import java.util.Scanner;

import org.ardverk.collection.PatriciaTrie;
import org.ardverk.collection.StringKeyAnalyzer;

public class Board {
	private Dice[][] field;
	private Dimension size;
	private boolean[][] history;

	/**
	 * Genere le plateau de des, a partir d'un fichier de definitions de ceux-ci.
	 * @param diceDefFileName Definition des des (un des par ligne, avec une chaine de 6 caracteres
	 * pour chaque face du des - les commentaires commencent par des point-virgules) 
	 * @throws FileNotFoundException
	 */
	public void setField(String diceDefFileName) throws FileNotFoundException {
		FileInputStream diceDefFile = new FileInputStream(diceDefFileName);
		Scanner diceDef = new Scanner(diceDefFile);
		ArrayList<String> diceDefs = new ArrayList<String>();
		int iDice = 0;

		field = new Dice[size.height][size.width];

		// Recuperation des faces des des depuis le fichier, en excluant les commentaires
		while (diceDef.hasNextLine()) {
			String line = diceDef.nextLine();
			
			if(!line.startsWith(";"))
				diceDefs.add(line);
		}

		// Creation des des, et du plateau qui les rassemble
		while (iDice < size.height * size.width) {
			int row = iDice / size.height;
			int col = iDice - row * size.width;

			field[row][col] = new Dice(diceDefs.get(iDice % diceDefs.size()).toCharArray());

			iDice++;
		}
	}
	
	/**
	 * Attribution d'un plateau de des deja cree (sert pour les tests unitaires).
	 * @param field Plateau de des
	 */
	public void setField(Dice[][] field) {
		this.field = field;
		
		size = new Dimension(
			field[0].length,
			field.length
		);
	}

	/**
	 * Melange des des du plateau : les faces de chacun des des ainsi que leur position sur le plateau
	 * sont choisies aleatoirement.
	 */
	public void roll() {
		Random rand = new Random();

		// Pour chaque des
		for (int row = 0; row < size.height; row++) {
			for (int col = 0; col < size.width; col++) {
				
				// Faire tourner le de
				field[row][col].roll();

				// Le positionner aleatoirement
				swap(row, col, row + rand.nextInt(size.height - row), col
						+ rand.nextInt(size.width - col));
			}
		}
	}

	/**
	 * Sortie texte.
	 */
	public String toString() {
		StringBuilder r = new StringBuilder(size.height * 3 * size.width * 5 + size.height * 3 * 2);

		for (int row = 0; row < size.height; row++) {
			r.append("\t");
			
			for (int col = 0; col < size.width; col++)
				r.append("/---\\");
			
			r.append("\n\t");
			
			for (int col = 0; col < size.width; col++)
				r.append("| ").append(field[row][col]).append(" |");
			
			r.append("\n\t");
			
			for (int col = 0; col < size.width; col++)
				r.append("\\---/");

			r.append("\n");
		}

		return r.toString();
	}

	/**
	 * Retourne la dimension du plateau de des
	 * @return Dimension du plateau de des 
	 */
	public Dimension getSize() {
		return size;
	}

	/**
	 * Imposition de la taille du plateau (les des disponibles seront repetes).
	 * @param width Taille (carre)
	 */
	public void setSize(int width) {
		size = new Dimension(width, width);
	}

	/**
	 * Retourne le plateau de des
	 * @return Plateau de des
	 */
	public Dice[][] getField() {
		return field;
	}

	/**
	 * Indique si un mot donne est lisible dans le plateau actuel.
	 * @param word Mot a tester
	 * @return Vrai si le mot est present sur le plateau
	 */
	public boolean isReadableInBoard(String word) {
		// valeur false par defaut dans java
		history = new boolean[size.height][size.width];
		boolean success = false;
		
		int row = 0, col = 0;

		// Pour chacune des lettres du plateau
		do {
			// L'emplacement du de est marque pour ne pas pouvoir etre reutilise (backtracking)
			history[row][col] = true;
			
			// Recherche le mot dans le plateau, en partant de la lettre donne et en voyageant
			// de proche en proche, en revenant en arriere chaque fois que le mot ne peut plus
			// etre genere
			if (isReadableInBoard_rec(
				word,
				row, col
			))
				success = true;
			
			// La recheche a echouee dans ce sens, le de est detaggue (retour arriere)
			else
				history[row][col] = false;

			if (++col >= size.width) {
				col = 0;
				row++;
			}

		// Tant qu'on a pas trouve le mot, ou qu'on a parcouru l'integralite du plateau
		} while(row < size.height && !success);

		return success;
	}

	/**
	 * Recherche le mot donne recursivement sur le plateau, en revenant en arriere quand il
	 * ne peut plus etre genere.
	 * @param searched Mot cherche (se reduit au fur et a mesure des occurences)
	 * @param row Ligne du dernier de lu 
	 * @param col Colonne du dernier de lu
	 * @return Vrai si le mot a ete trouve
	 */
	private boolean isReadableInBoard_rec(String searched, int row, int col) {
		boolean success = false;
		int i = 0;
		
		// Si la premiere lettre du bout de mot recherche ne correspond pas a la lettre
		// par laquelle on se propose de passer, on ne tente pas l'aventure (pour revenir
		// en arriere et trouver d'autres solutions)
		if(searched.charAt(0) == field[row][col].getValue()) {
			
			// Si il n'y a plus qu'une lettre, cela signifie que le mot dans sa totalite
			// a ete trouve : victoire et avortement
			if (searched.length() == 1)
				success = true;

			// Il reste des lettres a trouve pour avoir trouve le mot complet
			else { 

				// Toutes les directions que le backtracking peut explorer a partir du de
				// courant (null si la direction n'est pas possible)
				Point[] directions = lettersNearOf(col, row);
	
				// Pour chaque direction
				do {
					Point p = directions[i++];
					
					// Si la direction n'est pas possible, ou qu'elle a deja ete emprunte
					// dans le passe, on ne tente pas
					if(p != null && !history[p.x][p.y]) {
						history[p.x][p.y] = true;
	
						// Tentative de recherche des lettres restantes dans cette direction
						success = isReadableInBoard_rec(
							searched.substring(1),
							p.x, p.y
						);
	
						// Si cette direction s'est averee infructueuse, on note l'abandon en
						// indiquant que le de n'a finalement pas ete utilise (retour arriere)
						if(!success)
							history[p.x][p.y] = false;
					}
					
				// Tant que le mot n'a pas ete trouve, et qu'il reste des directions a explorer
				} while(!success && i < directions.length);
			}
		}

		return success;
	}
	
	/**
	 * Recherche tous les mots lisibles sur le plateau, a partir d'un dictionnaire.
	 * @param lexicon Dictionnaire
	 * @return Liste des mots trouves
	 */
	public PatriciaTrie<String, Object> allReadableInBoard(LexiconTrie lexicon) {
		history = new boolean[size.height][size.width];
		PatriciaTrie<String, Object> solutions = new PatriciaTrie<String, Object>(StringKeyAnalyzer.INSTANCE);

		int row = 0, col = 0;

		// Pour chaque de du plateau 
		do {
			// L'emplacement du de est marque pour ne pas pouvoir etre reutilise (backtracking)
			history[row][col] = true;

			// Recherche des mots dans le plateau, en partant de la lettre donne et en voyageant
			// de proche en proche, en revenant en arriere chaque fois qu'aucun mot ne peut etre
			// trouve a partir du prefixe forme
			allReadableInBoard_rec(
				String.valueOf(field[row][col].getValue()),
				row, col,
				lexicon,
				solutions
			);
			
			// La recherche dans ce sens a echoue, on detague le de
			history[row][col] = false;

			if (++col >= size.width) {
				col = 0;
				row++;
			}

		} while(row < size.height);
		
		return solutions;
	}
	
	/**
	 * Recherche recursivement tous les mots presents sur le plateau, a l'aide d'un dictionnaire.
	 * @param current Mot actuellement forme
	 * @param row Ligne du dernier de utilise
	 * @param col Colonne du dernier de utilise
	 * @param lexicon Dictionnaire
	 * @param solutions Liste des mots trouves
	 */
	private void allReadableInBoard_rec(String current, int row, int col, LexiconTrie lexicon, PatriciaTrie<String, Object> solutions) {

		// Abandon si le dictionnaire ne contient aucun mot ayant pour prefixe le mot
		// forme.
		if(lexicon.mayContain(current)) {

			// Ajout du mot forme dans les solutions si le dictionnaire le contient
			// On continue malgre tout a chercher d'autres mots ayant celui-ci pour
			// prefixe
			if (lexicon.contains(current))
				solutions.put(current, null);

			// Ensemble des directions possibles depuis le dernier de (null pour les
			// direction impossibles)
			Point[] directions = lettersNearOf(col, row);

			// Pour chaque direction
			for(Point p : directions) {
				
				// Si la direction est impossible ou que le prefixe forme est deja
				// passe par la, on abandonne (retour arriere)
				if(p != null && !history[p.x][p.y]) {
					
					// L'emplacement du de est marque, pour ne pas pouvoir repasser dessus
					history[p.x][p.y] = true;

					// Recherche d'un mot existant dans cette direction, avec le prefixe deja forme
					allReadableInBoard_rec(
						current + field[p.x][p.y].getValue(),
						p.x, p.y,
						lexicon,
						solutions
					);

					// La recherche dans cette direction a echouee : le de est detague
					history[p.x][p.y] = false;
				}
			}
		}
	}
	
	/**
	 * Inversion de deux des sur le plateau (sert pour le roll).
	 * @param row Ligne courante du premier de a inverser
	 * @param col Colonne courante du premier de a inverser
	 * @param rowDest Ligne courante du second de a inverser
	 * @param colDest Colonne courante du second de a inverser
	 */
	private void swap(int row, int col, int rowDest, int colDest) {
		Dice swaping = field[rowDest][colDest];

		// Inversion de place
		field[rowDest][colDest] = field[row][col];
		field[row][col] = swaping;
	}
	
	/**
	 * Toutes les directions possibles sur le plateau, depuis une position donne.
	 * Definie la direction a null si elle n'est pas possible (bord du plateau).
	 * TODO Permettre de rendre le plateau torique (facile) 
	 * @param col Colonne du de courant
	 * @param row Ligne du de courant
	 * @return
	 */
	private Point[] lettersNearOf(int col, int row) {
		return new Point[] {
            /* top      */ (row == 0 ? null : new Point(row-1, col)),
            /* topRight */ (row == 0 || col == size.width - 1 ? null : new Point(row-1, col+1)),
            /* right    */ (col == size.width - 1 ? null : new Point(row, col+1)),
            /* botRight */ (row == size.height - 1 || col == size.width - 1 ? null : new Point(row+1, col+1)), 
            /* bot      */ (row == size.height - 1 ? null : new Point(row+1, col)),
            /* botLeft  */ (row == size.height - 1 || col == 0 ? null : new Point(row+1, col-1)),
            /* left     */ (col == 0 ? null : new Point(row, col-1)),
            /* topLeft  */ (row == 0 || col == 0 ? null : new Point(row-1, col-1))
		};
	}
	
	/**
	 * Retourne le parcours utilise pour trouver le dernier mot cherche.
	 * (Utilise par IOGameGUI pour afficher les lettres correspondant au mot propose)
	 * @return Dernier parcours utilise
	 */
	public boolean[][] getHistory() {
		return history;
	}
}