/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.ArrayList;

public class Game {
	public static enum GAME_MODE { SOLO, MULTI, ONLINE_HOST, ONLINE_CLIENT };
	static final public int DEFAULT_SIZE = 4, TIMER_MAX = 3 * 60;
	private LexiconTrie lexicon;
	private Board board;
	private ArrayList<Player> players;
	private IOGame io;
	private NetworkClient clientSocket;
	private boolean acceptRemotePlayers, started;
	private GAME_MODE game_mode = null;
	private boolean interrupted;
	private int timer = 0;

	/**
	 * Cree un controleur de jeu. 
	 */
	public Game() {
		this.lexicon = new LexiconTrie();
		this.board = new Board();
		players = new ArrayList<Player>();
		io = new IOGameText(this);
	}
	
	/**
	 * Cree un controleur de jeu, avec un dictionnaire impose
	 * (utilise par les tests unitaires).
	 * @param lexicon Dictionnaire
	 */
	public Game(LexiconTrie lexicon) {
		this();
		this.lexicon = lexicon;
	}

	/**
	 * Definie la vue utilisee a partir des options passees au programme.
	 * Par defaut, c'est la vue texte qui est utilisee. 
	 * @param args Options du programme
	 */
	public void setIOFromArgs(String args[]) {
		if(args.length >= 1) {
			
			// Interface graphique
			if(args[0].equals("--gui"))
				setIO(new IOGameGUI(this));

			// Aide (quitte)
			if(args[0].equals("-h") || args[0].equals("--help")) {
				System.out.println("\n\t--gui : With GUI\n");
				System.exit(0);
			}
		}
	}
	
	/**
	 * Remet le jeu a zero (connexion perdue, utilisation du menu au cours d'une partie,
	 * nouveau jeu), en reinitialisant les valeurs et en stoppant les threads eventuellement
	 * lances.
	 */
	public void reset() {
		players = new ArrayList<Player>();
		started = false;
		game_mode = null;
		interrupted = false;
		timer = 0;

		NetworkMultiSocketServer.stopMain();
		NetworkMultiSocketServer.stopSynchro();
		NetworkSynchro.stopClient();
	}
	
	/**
	 * Appellee lorsqu'un joueur client pert la connexion avec le serveur. 
	 */
	public void lostConnection() {
		io.outErrorConnection();
		interrupt();
	}

	/**
	 * Permet a l'utilisateur de saisir autant de joueurs qu'il le souhaite.
	 * @throws InterruptedException
	 */
	public void addPlayers() throws InterruptedException {
		while(addPlayer());
	}
	
	/**
	 * Ajout du joueur client.
	 * @throws InterruptedException
	 */
	public void addPlayerClientLocal() throws InterruptedException {
		String name;
		
		name = io.inFirstPlayerName();
		
		// Demande tant que le nom n'est pas acceptable
		if(name != null && !name.isEmpty())
			addPlayer(new PlayerClientLocal(name, this));	
		else {
			io.outErrorPlayerRequired();
			addPlayerClientLocal();
		}
	}
	
	// Permet a l'utilisateur d'ajouter un joueur
	public boolean addPlayer() throws InterruptedException {
		String name;
		
		// Le premier joueur aura a definir la taille du plateau en meme temps
		// que de saisir son nom
		if(players.isEmpty())
			name = io.inFirstPlayerNameWithFieldWidth();
		else
			name = io.inPlayerName();

		// Nom deja utilise : on redemande sans ajouter
		if(hasPlayerName(name)) {
			io.outErrorPlayerName();
			return addPlayer();
		}

		// Ajout du joueur, si le nom n'est pas vide (volonte de ne plus saisir de joueur,
		// ce qui est possible uniquement si un joueur au minimum a deja ete renseigne)
		if(name != null && !name.isEmpty()) {
			addPlayer(new PlayerHuman(name, this));
			return true;
		}

		// Dans le cas contraire, on signale qu'il faut un joueur minimum et on le redemande
		if(players.isEmpty()) {
			io.outErrorPlayerRequired();
			return addPlayer();
		}

		return false;
	}
	
	/**
	 * Debut de la collecte des joueurs distants pour le serveur.
	 */
	public void authorizeRemotePlayers() {
		acceptRemotePlayers = true;
	}
	
	/**
	 * Fin de la collecte des joueurs distants pour le serveur.
	 */
	public void noNewRemotePlayers() {
		acceptRemotePlayers = false;
	}
	
	/**
	 * Debut de la partie.
	 */
	public void start() {
		started = true;
	}
	
	/**
	 * Ajout du robot aux joueurs de la partie.
	 */
	public void addBotPlayer() {
		addPlayer(new PlayerBot(this));
	}
	
	/**
	 * Fait jouer le robot (trouve tous les mots qui n'ont pas ete trouves
	 * par les joueurs.
	 */
	public void playBotPlayer() {
		for(Player p : players)
			if(p instanceof PlayerBot)
				((PlayerBot) p).play();
	}

	/**
	 * Ajout d'un joueur.
	 * @param p Joueur
	 */
	public void addPlayer(Player p) {
		players.add(p);
	}
	
	/**
	 * Renvoie une WordException si le mot indique est plus petit que la taille minimal autorisee.
	 * @param word Mot a tester
	 * @throws WordException
	 */
	public void minCheckWord(String word) throws WordException {
		if(word.length() < 3) {
			throw new WordException("Words shorter than 3 chars are not allowed !");
		}
	}
	
	/**
	 * Teste le mot donne, et renvoie une WordException si celui-ci n'est pas lisible
	 * dans le plateau, ou n'est pas dans le dictionnaire. 
	 * @param word Mot a tester
	 * @throws WordException
	 */
	public void fullCheckWord(String word) throws WordException {
		minCheckWord(word);
		
		// Le mot n'est pas lisible
		if(!board.isReadableInBoard(word))
			throw new WordException("This word is not in the board.");

		// Le mot n'est pas dans le dictionnaire
		if(!lexicon.contains(word))
			throw new WordException("This word is not in the lexicon.");
	}
	
	/**
	 * Calcul le score d'un mot donne.
	 * @param word Mot a evaluer
	 * @return Score du mot
	 */
	public int wordScore(String word) {
		switch(word.length()) {
			case 3 :
			case 4 :
				return 1;
			case 5 :
				return 2;
			case 6 :
				return 3;
			case 7 :
				return 5;
			default :
				return 11;				
		}
	}

	/**
	 * Retourne la vue utilisee par le controleur.
	 * @return Vue utilisee par le controleur.
	 */
	public IOGame getIO() {
		return io;
	}
	
	/**
	 * Permet de definir la vue a utiliser.
	 * @param io Vue a utiliser
	 */
	public void setIO(IOGame io) {
		this.io = io;
	}
	
	/**
	 * Retourne le dictionnaire utilise.
	 * @return Dictionnaire utilise
	 */
	public LexiconTrie getLexicon() {
		return lexicon;
	}
	
	/**
	 * Retourne le plateau utilise. 
	 * @return Plateau utilise
	 */
	public Board getBoard() {
		return board;
	}
	
	/**
	 * Retourne l'etat de l'autorisation de la collecte de joueurs distants pour le serveur.
	 * @return Vrai si la collecte est en cours
	 */
	public boolean getAcceptRemotePlayers() {
		return acceptRemotePlayers;
	}
	
	/**
	 * Retourne l'etat de la partie.
	 * @return Vrai si une partie est en cours
	 */
	public boolean isStarted() {
		return started;
	}

	/**
	 * Interrompt la partie en cours.
	 */
	public void interrupt() {
		interrupted = true;
	}
	
	/**
	 * Retourne si la derniere partie a ete interrompue.
	 * @return Vrai si la derniere partie a ete interrompue
	 */
	public boolean isInterrupted() {
		return interrupted;
	}

	/** 
	 * Melange le plateau.
	 */
	public void mixup() {
		board.roll();
	}

	/**
	 * Fait jouer les joueurs enregistres.
	 * @throws IOException
	 * @throws InterruptedException
	 */
	public void playing() throws IOException, InterruptedException {
		io.outBoard();
		io.outUpdateScores();

		// Pour chaque jour enregistre
		for(Player p : players) {
			
			// Seuls les joueurs physiques sur la machine peuvent jouer
			if(!(p instanceof PlayerRemote || p instanceof PlayerClientRemote)) {
				if(p instanceof PlayerHuman) {
					io.outPlayerPlaying(p.getName());
					p.play();
				}
			}
		}
	}
	
	/**
	 * Retourne la liste des joueurs de la (prochaine) partie.
	 * @return Liste des joueurs de la (prochaine) partie
	 */
	public ArrayList<Player> getPlayers() {
		return players;
	}
	
	/**
	 * Lance en tache de fond le serveur qui ouvrira un socket pour creer un thread pour chaque
	 * joueur qui souhaite rejoindre la partie. Envoi une InterruptedException si la partie s'interrompt
	 * brutalement.
	 * @throws InterruptedException
	 */
	public void startOnlineServer() throws InterruptedException {
		
		// Ouvre le socket principal pour accepter pour communiquer avec les joueurs
		// distants
		NetworkMultiSocketServer.startMain(this);

		// Autorise la collecte des joueurs distants
		authorizeRemotePlayers();

		// Ferme l'inscription de nouveau joueurs quand l'utilisateur (du serveur) le decide
		if(io.inNoNewRemotePlayers())
			noNewRemotePlayers();

		// Lance en tache de fond le serveur qui s'occupera de lancer des threads dedies
		// a la synchro des scores, pour chacun des utilisateurs distants
		NetworkMultiSocketServer.startSynchro(this);
	}
	
	/**
	 * Retourne le joueur correspondant a une identite donnee
	 * @param identity Identite d'un joueur (IP + SEPARATEUR + nom)
	 * @return Joueur correspondant
	 */
	public PlayerRemote getRemotePlayerByIdentity(String identity) {
		for(Player p : players)
			if(p instanceof PlayerRemote)
				if(((PlayerRemote) p).getRemoteIdentity().equals(identity))
					return (PlayerRemote) p;

		return null;
	}

	/**
	 * Retourne le joueur correspondant a une IP donnee.
	 * TODO Pose un gros probleme si deux joueurs sont derriere un meme NAT
	 * @param address IP du joueur
	 * @return Joueur correspondant
	 */
	public PlayerRemote getRemotePlayerByAddress(String address) {
		for(Player p : players) {
			if(p instanceof PlayerRemote)
				if(((PlayerRemote) p).getAddress().equals(address))
					return (PlayerRemote) p;
		}

		return null;
	}

	/**
	 * Desactive un joueur distant (cas d'un joueur soudainement injoignable en cours de partie).
	 * Il reste dans les joueurs, mais ses threads reseaux sont coupes.
	 * @param rp Joueur injoignable
	 * @param thread Runnable threade du joueur
	 */
	public void disableRemotePlayer(PlayerRemote rp, Network thread) {
		rp.disable();
		
		// La coupure a ete detecte lors d'une tentative d'envoi ou de reception
		// d'informations
		if(thread instanceof NetworkServer)
			((NetworkServer) thread).stop();
		
		// La coupure a ete detectee lors d'une tentative de synchro des scores
		if(thread instanceof NetworkSynchro)
			((NetworkSynchro) thread).stop();
		
		io.outUpdateScores();
	}

	/**
	 * Indique si un joueur porte deja le nom donne.
	 * @param name Nom a tester
	 * @return Vrai si un joueur porte deja ce nom
	 */
	public boolean hasPlayerName(String name) {
		for(Player p : players)
			if(p.getName().equals(name))
				return true;
		
		return false;
	}

	/**
	 * Retourne l'instance utilisee pour la communication avec le serveur,
	 * par le client.
	 * @return Instance utilisee pour la communication avec le serveur
	 */
	public NetworkClient getClientSocket() {
		return clientSocket;
	}
	
	/**
	 * Permet de creer l'instance utilisee pour la communication avec le serveur.
	 * @throws UnknownHostException
	 * @throws IOException
	 */
	public void setClientSocket() throws UnknownHostException, IOException {
		this.clientSocket = new NetworkClient(this);
	}
	
	/**
	 * Indique si la partie est terminee (tous les joueurs ont fini de jouer).
	 * @return Vrai si tous les joueurs ont fini de jouer
	 */
	public boolean isFinish() {
		
		// Si le chrono a expire, on considere que la partie est forcement finie
		if(timer < TIMER_MAX) {
			for(Player p : players)
				if(p instanceof PlayerHuman && !(p instanceof PlayerClientRemote) && !p.isFinish())
					return false;
		}

		return true;
	}

	/**
	 * Retourne le mode de jeu actuellement choisi (solo, multi ou online).
	 * @return Mode de jeu actuellement choisi (solo, multi ou online)
	 */
	public GAME_MODE getGameMode() {
		return game_mode;
	}

	/**
	 * Permet de definir le mode de jeu souhaite.
	 * @param game_mode Mode de jeu souhaite
	 */
	public void setGameMode(GAME_MODE game_mode) {
		this.game_mode = game_mode;
	}

	/**
	 * Retourne l'avancement actuel du chronometre.
	 * @return Avancement actuel du chronometre
	 */
	public int getTimer() {
		return timer;
	}
	
	/**
	 * Incremente le chronometre de 1 seconde.
	 */
	public void incrementTimer() {
		timer++;
	}
}