/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.Timer;
import javax.swing.*;
import javax.swing.event.*;

import boggle.Game.GAME_MODE;

public class IOGameGUI extends IOGame {
	private static final long serialVersionUID = -2473540531915361158L;
	private JPanel mainPan, boardPan, playersPan, playPan;
	private JTextField wordField;
	private JLabel playerLabel;
	private JMenuBar menu;
	private JButton buttonField, buttonDone;
	private boolean proposedWord;
	private JProgressBar loading, timer;
	private String selectedWord = null, player_name;
	private Game.GAME_MODE game_mode = null;
	private JLabel[][] letters;
	private HashMap<String, boolean[][]> historiesBuffer = new HashMap<String, boolean[][]>();

	// Couleurs de l'interface
	final Color BOARD_COLOR = Color.decode("#cccccc"),
		PLAYERSPAN_COLOR = Color.decode("#757575"),
		DICE_COLOR = Color.decode("#c8c8c8"),
		DICE_BORDER_COLOR = Color.decode("#525252"),
		SELECTED_DICE_COLOR = Color.decode("#b55e5e"),
		TITLE_COLOR = Color.decode("#eadeb4"),
		PLAYERS_COLOR = Color.decode("#EEEEEE"),
		CORRECT_COLOR = Color.decode("#29b023"),
		ERROR_COLOR = Color.decode("#ed273d"),
		LIST_COLOR = Color.decode("#dddddd");

	final Font TITLE_FONT = new Font("Arial", Font.BOLD, 19);
	
	/**
	 * Cree une vue graphique.
	 * @doc http://download.oracle.com/javase/tutorial/uiswing/components/menu.html
	 * @param game Controleur du jeu
	 */
	public IOGameGUI(Game game) {
		super(game);
		final Game g = game;

		setTitle("Boggle");

		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();

        // Fenetre principale
		mainPan = new JPanel();
        mainPan.setLayout(new BorderLayout());

        // Les differents panels
        boardPan = new JPanel();
        playPan = new JPanel();
        playersPan = new JPanel();
        menu = new JMenuBar();
        
        boardPan.setBackground(BOARD_COLOR);

        // Menu principal
        JMenu subMenuPlay = new JMenu("Nouveau");
        JMenu subMenuHelp = new JMenu("Aide");

        menu.add(subMenuPlay);
        menu.add(subMenuHelp);

        // Creation des sous-menus
        JMenuItem subMenuPlaySolo = new JMenuItem("Jeu solo");
        JMenuItem subMenuPlayMulti = new JMenuItem("Jeu multijoueurs");
        JMenu subMenuPlayOnline = new JMenu("Jeu en ligne");
        JMenuItem subMenuPlayOnlineHost = new JMenuItem("Héberger");
        JMenuItem subMenuPlayOnlineClient = new JMenuItem("Rejoindre");
        JMenuItem subMenuHelpAbout = new JMenuItem("À propos");
        JMenuItem subMenuHelpQuit = new JMenuItem("Quitter");

        subMenuPlay.add(subMenuPlaySolo);
        subMenuPlay.add(subMenuPlayMulti);
        subMenuPlay.add(subMenuPlayOnline);
        subMenuPlayOnline.add(subMenuPlayOnlineHost);
        subMenuPlayOnline.add(subMenuPlayOnlineClient);
        subMenuHelp.add(subMenuHelpAbout);
        subMenuHelp.add(new JSeparator());
        subMenuHelp.add(subMenuHelpQuit);

        // Jeu en solo
        subMenuPlaySolo.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				changeGameMode(GAME_MODE.SOLO);
			}
		});
		
        // Jeu en multijoueurs (sur la meme machine)
        subMenuPlayMulti.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				changeGameMode(GAME_MODE.MULTI);
			}
		});

        // Jeu en ligne, en hebergeant
        subMenuPlayOnlineHost.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				changeGameMode(GAME_MODE.ONLINE_HOST);
			}
		});
        
        // Jeu en ligne, en rejoignant un serveur 
        subMenuPlayOnlineClient.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				changeGameMode(GAME_MODE.ONLINE_CLIENT);
			}
		});

        // A propos
        subMenuHelpAbout.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				about();
			}
		});
        
        // Quitter
        subMenuHelpQuit.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				if(g.isStarted()) {
					if(inStop())
						System.exit(0);
				} else
					System.exit(0);
			}
		});

        // Creation de la structure
        playPan.setLayout(new BoxLayout(playPan, BoxLayout.X_AXIS));
        playersPan.setLayout(new BoxLayout(playersPan, BoxLayout.Y_AXIS));

	    playersPan.setPreferredSize(new Dimension(230, 100));
	    playersPan.setBackground(PLAYERSPAN_COLOR);
	    playersPan.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
	    playersPan.setVisible(false);

	    setJMenuBar(menu);
	    mainPan.add(boardPan, BorderLayout.CENTER);
	    mainPan.add(playersPan, BorderLayout.EAST);
	    mainPan.add(playPan, BorderLayout.SOUTH);

        playerLabel = new JLabel();
        wordField = new JTextField();
        buttonField = new JButton("Proposer");
        buttonDone = new JButton("Terminé");

        // Champ d'ajout des mots
        wordField.setFont(new Font("Arrial", Font.BOLD, 13));
        wordField.setHorizontalAlignment(JTextField.CENTER);

        wordField.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent ae) {
        		proposedWord = true;
        	}
        });

        buttonField.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent ae) {
        		proposedWord = true;
        	}
        });
        
        buttonDone.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent ae) {
        		wordField.setText("");
        		proposedWord = true;
        	}
        });

        timer = new JProgressBar(0, 100);
		timer.setForeground(ERROR_COLOR);

        getContentPane().add(mainPan);
        setVisible(true);
        setSize(screenSize.width / 100 * 60, screenSize.height / 100 * 60);

        // Fenetre centree (sauf si tiling qui rox, comme xmonad)
        setLocation(
                Math.max(0, (screenSize.width  - getSize().width) / 2),
                Math.max(0, (screenSize.height - getSize().height) / 2)
        );

        setDefaultCloseOperation(EXIT_ON_CLOSE);
	}

	@Override
	public void outBeginWaitLexicon() {
		menu.setEnabled(false);

		loading = new JProgressBar();
		loading.setIndeterminate(true);

		playPan.add(loading);

		validate();
		repaint();
	}

	@Override
	public void outEndWaitLexicon() {
		playPan.remove(loading);
	}

	@Override
	public boolean inNoNewRemotePlayers() throws InterruptedException {
		final Game g = game;
		JButton button = new JButton("Lancer la partie");

		// Possibilite d'interrompre la partie a ce stade
		menu.setEnabled(true);
	    playersPan.setVisible(true);

	    // Mise a jour du panneau des joueurs entrants, tant que la collecte a lieu 
		while(game.getAcceptRemotePlayers()) {
			playersPan.removeAll();
			
			// Affichage des joueurs
			for(Player p : game.getPlayers()) {
				if(!(p instanceof PlayerBot)) {
					JLabel n = new JLabel(p.toString() + " a rejoint la partie.");
					n.setForeground(PLAYERS_COLOR);
					playersPan.add(n);
				}
			}

			// Bouton de lancement de la partie
			button.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					g.noNewRemotePlayers();
				}
			});

			playersPan.add(new JLabel(" "));
			playersPan.add(button);
			
			validate();
			repaint();

			// Partie interrompue (l'utilisateur a utilise le menu)
			if(game.isInterrupted())
				throw new InterruptedException();

			try {
				Thread.sleep(1000);
			} catch (InterruptedException e1) {}
		}
		
		return true;
	}

	@Override
	public String inHostAddress() {
		menu.setEnabled(false);
		
		Network.setHost(JOptionPane.showInputDialog(
			this,
			"IP de l'hôte : ",
			"Hôte",
			JOptionPane.QUESTION_MESSAGE
		));
		
		return null;
	}

	@Override
	public String inPlayerName() {
		Object[] options = { "Ajouter", "Terminé" };

		JOptionPane dialogPan = new JOptionPane(
				"Nom du joueur :",
				JOptionPane.PLAIN_MESSAGE,
				2, UIManager.getIcon("OptionPane.questionIcon"),
				options,
				null
		);

		dialogPan.setWantsInput(true);
		JDialog dialog = dialogPan.createDialog(this, "Ajouter un joueur");

		dialogPan.selectInitialValue();
		dialog.setVisible(true);
		dialog.dispose();
		
		if(dialogPan.getValue() == null || dialogPan.getValue() == options[1])
			return "";
		else
			return (String) dialogPan.getInputValue();
	}

	@Override
	public String inFirstPlayerName() throws InterruptedException {
		Object[] options = { "Rejoindre", "Annuler" };
		
		JOptionPane dialogPan = new JOptionPane(
			"Nom du joueur :",
			JOptionPane.PLAIN_MESSAGE,
			2, UIManager.getIcon("OptionPane.questionIcon"),
			options,
			null
		);

		dialogPan.setWantsInput(true);
		dialogPan.setInputValue(player_name);
		JDialog dialog = dialogPan.createDialog(this, "Rejoindre la partie");
		
		dialog.setVisible(true);
		dialog.dispose();

		if(dialogPan.getValue() == null || dialogPan.getValue() == options[1])
			throw new InterruptedException();

		player_name = (String) dialogPan.getInputValue();
		return player_name;
	}
	
	@Override
	public String inFirstPlayerNameWithFieldWidth() throws InterruptedException {
		Object[] options = { game.getGameMode() == Game.GAME_MODE.SOLO ? "Jouer" : "Ajouter", "Annuler" };
        JSlider fieldWidth = new JSlider(JSlider.HORIZONTAL, 3, 20, Game.DEFAULT_SIZE);

		JComponent[] inputs  = {
			new JLabel("Largeur du plateau :"),
			fieldWidth,
			new JLabel(" "),
			new JLabel("Nom du joueur :"),
		};

		fieldWidth.setMajorTickSpacing(5);
		fieldWidth.setMinorTickSpacing(1);
		fieldWidth.setPaintTicks(true);
		fieldWidth.setPaintLabels(true);
        fieldWidth.setSnapToTicks(true);

        game.getBoard().setSize(Game.DEFAULT_SIZE);
        
		fieldWidth.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent e) {
				game.getBoard().setSize(((JSlider) e.getSource()).getValue());
			}
		});
        
		JOptionPane dialogPan = new JOptionPane(
			inputs,
			JOptionPane.PLAIN_MESSAGE,
			2, UIManager.getIcon("OptionPane.questionIcon"),
			options,
			null
		);

		dialogPan.setWantsInput(true);
		dialogPan.setInputValue(player_name);
		JDialog dialog = dialogPan.createDialog(this, "Lancement du jeu");

		dialog.setVisible(true);
		dialog.dispose();

		if(dialogPan.getValue() == null || dialogPan.getValue() == options[1])
			throw new InterruptedException();

		player_name = (String) dialogPan.getInputValue();
		return player_name;
	}

	@Override
	public String inWord() {
		menu.setEnabled(true);
		wordField.requestFocus();
			
		// Le champ de texte indique lui-meme quand l'utilisateur a valide un mot
		// en changeant le flag de valeur
		while(!proposedWord) {
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {}
			
			if(game.isInterrupted()) {
				wordField.setText("");
				proposedWord = true;
			}
		}

		String word = wordField.getText();
		wordField.setText("");
		proposedWord = false;
		
		// Si le mot est vide
		if(word.isEmpty()) {
			
			// Desactivation du champ de texte si on est pas en multi
			if(game.getGameMode() != Game.GAME_MODE.MULTI)
				wordField.setEditable(false);
		
			// Si c'est du en ligne on indique qu'il va falloir attendre que les autres finissent pour avoir les resultats
			if((game.getGameMode() == Game.GAME_MODE.ONLINE_HOST || game.getGameMode() == Game.GAME_MODE.ONLINE_CLIENT) && game.getTimer() < Game.TIMER_MAX)
				outInfo("Validation enregistrée", "Les résultats vont s'afficher.\n\nIl reste du temps, et d'autres joueurs n'ont pas encore fini de jouer.\nMerci de patienter.");
		}

		// Commentaire absent
		if(word.equals("42"))
			outEE();

		return word;
	}
	
	/**
	 * Retabli les des marques, apres avoir demande de tracer le chemin d'un mot sur le plateau
	 */
	private void noViewWord() {
		for(int row = 0; row < letters.length; row++)
			for(int col = 0; col < letters[0].length; col++)
				letters[row][col].setBackground(DICE_COLOR);
		
		boardPan.validate();
		boardPan.repaint();
		
		selectedWord = null;
	}
	
	/**
	 * Met en valeur les lettres qui permettent de former un mot, indique dans les resultats.
	 */
	private void viewWord() {
		boolean[][] fieldTrace;
		String word = new String(selectedWord);

		// Cache des traces deja recherches
		if(historiesBuffer.containsKey(word))
			fieldTrace = historiesBuffer.get(word);
		else {
			
			// Demande d'un trace, en recuperant l'historique du Game, apres avoir
			// demande si un mot est dans le plateau
			game.getBoard().isReadableInBoard(word);
			fieldTrace = game.getBoard().getHistory();
			
			// Ajout au cache
			historiesBuffer.put(word, fieldTrace);
		}

		// Mise en valeur des des concernes
		for(int row = 0; row < fieldTrace.length; row++)
			for(int col = 0; col < fieldTrace[0].length; col++)
				letters[row][col].setBackground(fieldTrace[row][col] ? SELECTED_DICE_COLOR : DICE_COLOR);

		boardPan.validate();
		boardPan.repaint();

		validate();
		repaint();

		// Attend qu'un nouveau mot soit selectionne, pour le mettre en valeur sur le plateau
		while(selectedWord != null && selectedWord.equals(word))
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {}

		// Retour a la boucle
		inReplay();
	}

	@Override
	public void inReplay() {
		
		// Ne decroche pas tant qu'un nouveau jeu n'a pas ete demande, ou que l'affichage
		// d'un mot n'a pas ete reclame
		while(game.getGameMode() == null && selectedWord == null)
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {}

		// Si on a decroche parce qu'on voulait un mot, on l'affiche (et on reviendra)
		if(selectedWord != null)
			viewWord();
		
		// Sinon c'est fini, un nouveau jeu va demarrer
		else
			reset();
	}

	/**
	 * Affiche un popup d'informations.
	 * @param title Titre du popup
	 * @param msg Message du popup
	 */
	private void outInfo(String title, String msg) {
		JOptionPane.showMessageDialog(
			this,
			msg,
			title,
			JOptionPane.INFORMATION_MESSAGE
		);
	}
	
	/**
	 * Affiche un popup d'erreur.
	 * @param title Titre du popup
	 * @param msg Message du popup
	 */
	private void outError(String title, String msg) {
		JOptionPane.showMessageDialog(
			this,
			msg,
			title,
			JOptionPane.ERROR_MESSAGE
		);
	}
	
	@Override
	public void outErrorPlayerRequired() {
		outError("Nom requis", "Vous devez saisir un nom de joueur.");
	}
	
	@Override
	public void outErrorPlayerName() {
		outError("Nom invalide", "Un autre joueur porte déjà ce nom.");
	}
	
	@Override
	public void outErrorConnection() {
		outError("Serveur injoignable", "Le serveur est injoignable.");
	}

	@Override
	public void outErrorWord(String msg) {
		outInline(msg, true);
	}

	@Override
	public void outBanner() {
		JLabel bannerLabel, subBannerLabel;
		
		bannerLabel = new JLabel("Boggle", JLabel.CENTER);
		bannerLabel.setFont(new Font("Arial", Font.BOLD,90));
		bannerLabel.setForeground(TITLE_COLOR);
		
		subBannerLabel = new JLabel("TOP PROJET");
		subBannerLabel.setAlignmentY(JLabel.SOUTH_EAST);
		subBannerLabel.setFont(new Font("Arial", Font.BOLD, 10));

		boardPan.setLayout(new GridBagLayout());
		boardPan.add(bannerLabel);
		boardPan.add(subBannerLabel);
	}
	
	@Override
	public void inGameMode() {
		if(game_mode != null) {
			changeGameMode(game_mode);
			game_mode = null;
		} else
			while(game.getGameMode() == null)
				try {
					Thread.sleep(200);
				} catch (InterruptedException e) {}
	}
	
	@Override
	public void outBeginWaitServer() {
		loading = new JProgressBar();
		loading.setIndeterminate(true);

		playPan.add(loading);

		outInfo("Attente du server", "Vous avez été enregistré.\nMerci de patienter pour rejoindre la partie.");

		validate();
		repaint();
	}

	@Override
	public void outEndWaitServer() {
		playPan.remove(loading);
	}
	
	@Override
	public void outPlayerPlaying(String name) {
		
		// Indique qui joue, en mode multi
		if(game.getGameMode() == Game.GAME_MODE.MULTI) {
			timer.setVisible(false);
			outInfo("Changement de joueur", "Joueur suivant : " + name);
		}

		playerLabel.setText(
			(game.getGameMode() == Game.GAME_MODE.MULTI ? " [" + name + "]" : "") +
			" Proposer un mot : "
		);
	}
	
	@Override
	public void outCorrectWord() {
		outInline("CORRECT", false);
	}
	
	@Override
	public void outResults() {
		final HashMap<String, ArrayList<Player>> words = new HashMap<String, ArrayList<Player>>();
		final ArrayList<String> wordsKeys = new ArrayList<String>();
		final JList wordsList;
		DefaultListModel wordsListModel;
		JScrollPane wordsListScroller;
		
		// Cache la barre de jeu
		playPan.setVisible(false);
		
		// Pour chacun des joueurs, on recolte ses mots
		for(Player p : game.getPlayers())
			if(p instanceof PlayerHuman)
				
				// Pour chacun de ses mots, on ajoute dans une liste, en l'associant avec lui
				for(String word : p.getWords().keySet())
					if(!words.containsKey(word)) {
						words.put(word, new ArrayList<Player>());
						words.get(word).add(p);
						wordsKeys.add(word);

						// Si le mot a ete trouve par qqn d'autre, on associe aussi ce joueur
						if(!p.getOriginalWords().containsKey(word))
							for(Player p2 : game.getPlayers())
								if(p2 instanceof PlayerHuman && p != p2 && p2.getWords().containsKey(word))
									words.get(word).add(p2);
					}
		
		// Trie des mots
		Collections.sort(wordsKeys);

		// Titre
		JLabel title = new JLabel("MOTS TROUVÉS");
		title.setFont(TITLE_FONT);
		title.setForeground(TITLE_COLOR);
		playersPan.add(new JLabel(" "));
		playersPan.add(title);

		wordsListModel = new DefaultListModel();
		
		for(String word : wordsKeys)
			wordsListModel.addElement(word);

		// Liste des mots trouves par les joueurs
		wordsList = new JList(wordsListModel) {
			private static final long serialVersionUID = -1947311138060989771L;

			// Affiche une infobulle sur les mots qui indique qui l'a trouve
			public String getToolTipText(MouseEvent e) {
				int index = locationToIndex(e.getPoint());

				if (-1 < index) {
					String tooltip = "";
					
					for(Player p : words.get((String) getModel().getElementAt(index)))
						tooltip += p + " ";
					
					return tooltip;

				} else
					return null;
			}
		};
		
		wordsList.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
		wordsList.setLayoutOrientation(JList.HORIZONTAL_WRAP);
		wordsList.setVisibleRowCount(-1);
		wordsList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		wordsList.setPreferredSize(new Dimension(190, 100));

		// Permet de faire surligner le mot si on clique dessus
		wordsList.addListSelectionListener(new ListSelectionListener() {
			
			@Override
			public void valueChanged(ListSelectionEvent e) {
				if(!wordsList.isSelectionEmpty())
					selectedWord = (String) wordsKeys.get(wordsList.getSelectedIndex());
			}
		});

		wordsListScroller = new JScrollPane(wordsList);
		wordsList.setSelectionBackground(SELECTED_DICE_COLOR);
		wordsList.setBackground(LIST_COLOR);

		// Permet de faire desurligner le mot, si la souris quitte la liste
		wordsList.addMouseListener(new MouseListener() {
			@Override public void mouseReleased(MouseEvent e) {}
			@Override public void mouseClicked(MouseEvent e) {}
			@Override public void mousePressed(MouseEvent e) {}
			@Override public void mouseEntered(MouseEvent e) {}

			@Override
			public void mouseExited(MouseEvent e) {
				wordsList.clearSelection();
				noViewWord();
			}
		});

		playersPan.add(wordsListScroller);
		
		// Titre
		title = new JLabel("IL Y AVAIT AUSSI");
		title.setFont(TITLE_FONT);
		title.setForeground(TITLE_COLOR);
		playersPan.add(new JLabel(" "));
		playersPan.add(title);

		// Recherche des mots du robot (non trouves)
		for(Player p : game.getPlayers())
			if(p.getName().equals(PlayerBot.DEFAULT_NAME)) {
				final JList wordsBotList;
				JScrollPane wordsBotListScroller; 

				// Creation de la liste des mots
				wordsBotList = new JList(p.getOriginalWords().keySet().toArray());
				wordsBotList.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
				wordsBotList.setLayoutOrientation(JList.HORIZONTAL_WRAP);
				wordsBotList.setVisibleRowCount(-1);
				wordsBotList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
				wordsList.setPreferredSize(new Dimension(190, 100));

				// Permet de mettre en evidence le mot dans le plateau si on clique dessus
				wordsBotList.addListSelectionListener(new ListSelectionListener() {
					@Override
					public void valueChanged(ListSelectionEvent e) {
						if(!wordsBotList.isSelectionEmpty())
							selectedWord = (String) wordsBotList.getSelectedValue();
					}
				});

				wordsBotListScroller = new JScrollPane(wordsBotList);
				wordsBotList.setSelectionBackground(SELECTED_DICE_COLOR);
				wordsBotList.setBackground(LIST_COLOR);

				// Permet de desurligner le mot, si la souris quitte la liste
				wordsBotList.addMouseListener(new MouseListener() {
					@Override public void mouseReleased(MouseEvent e) {}
					@Override public void mouseClicked(MouseEvent e) {}
					@Override public void mousePressed(MouseEvent e) {}
					@Override public void mouseEntered(MouseEvent e) {}

					@Override
					public void mouseExited(MouseEvent e) {
						wordsBotList.clearSelection();
						noViewWord();
					}
				});

				playersPan.add(wordsBotListScroller);
			}
				
		validate();
		repaint();				
	}
	
	@Override
	public void outElapsedTime() {
		outInfo("Temps écoulé", "La partie est finie.");
	}

	@Override
	public void outBoard() {
	    playersPan.setVisible(true);
		
		letters = new JLabel[game.getBoard().getSize().width][game.getBoard().getSize().height];
		boardPan.removeAll();

		// Creation de la grille du plateau
		boardPan.setLayout(
			new GridLayout(
				game.getBoard().getSize().width,
				game.getBoard().getSize().height
			)
		);

		Dice[][] field = game.getBoard().getField();
		
		// On ajoute tous les des au plateau
		for(int row = 0; row < letters.length; row++)
			for(int col = 0; col < letters[0].length; col++) {
				JLabel c = new JLabel(field[row][col].toString(), JLabel.CENTER);
				
				c.setFont(new Font("Arial", Font.BOLD, 35));
				c.setBorder(BorderFactory.createLineBorder(DICE_BORDER_COLOR, 1));
				c.setBackground(DICE_COLOR);
				c.setOpaque(true);
				
				letters[row][col] = c;
				boardPan.add(c);
			}

	    playPan.add(new JLabel(" "));
	    playPan.add(timer);
	    playPan.add(playerLabel);
	    playPan.add(wordField);
	    playPan.add(buttonField);
	    playPan.add(buttonDone);
		
		validate();
		repaint();
	}
	
	/**
	 * Affiche un message directement dans le prompt.
	 * @source Repris de Franklin
	 * @param msg Message a afficher
	 * @param error Vrai si il s'agit d'un message d'erreur
	 */
	private void outInline(String msg, boolean error) {
		try {
			wordField.setEditable(false);
			wordField.setBackground(error ? ERROR_COLOR : CORRECT_COLOR);
			wordField.setForeground(Color.WHITE);
			wordField.setText(msg.toUpperCase());
	
			Timer blink = new Timer();
	
			// Efface le message au bout de qq secondes 
			TimerTask blinking = new TimerTask() {
				public void run() {
					wordField.setBackground(Color.WHITE);
					wordField.setForeground(Color.BLACK);
					wordField.setText("");
					wordField.setEditable(true);
				}
			};

			// Lance le clignotement
			GregorianCalendar stopBlink = new GregorianCalendar();
			stopBlink.add(GregorianCalendar.MILLISECOND, 700);
			blink.schedule(blinking, stopBlink.getTime());

		} catch(NullPointerException e)  {}
	}

	@Override
	public void outUpdateTimer() {
		try {
			// Le chrono est indique avec une barre de chargement
			timer.setValue(Math.round((float) game.getTimer() * 100 / Game.TIMER_MAX));
			timer.setToolTipText(timer.getValue() + "% des " + (Game.TIMER_MAX / 60) + "  minutes est écoulé");
	
			validate();
			repaint();

		} catch(NullPointerException e)  {}
	}
	
	@Override
	@SuppressWarnings("unchecked")
	public void outUpdateScores() {
		ArrayList<Player> players;

		try {
			playersPan.removeAll();
			
			// Clone de la liste des joueurs (pour ne pas l'affecte au tri)
			players = (ArrayList<Player>) game.getPlayers().clone();
	
			// Titre
			JLabel title = new JLabel("SCORES");
			title.setFont(TITLE_FONT);
			title.setForeground(TITLE_COLOR);
			playersPan.add(title);
			playersPan.add(new JLabel(" "));
	
			// Tri des joueurs (en fonction de leur score)
			Collections.sort(players, Collections.reverseOrder());
	
			// Pour chaque joueur
			for(Player p : players)
				if(!(p instanceof PlayerBot)) {
					int score = p.getScore();
					JLabel s = new JLabel("# " + p.toString() + " : " + score + " point" + (score > 1 ? "s" : ""));
	
					// Affichage du joueur en rouge, si il n'est plus connecte (serveur)
					if(p instanceof PlayerRemote && ((PlayerRemote) p).isDisabled())
						s.setForeground(ERROR_COLOR);
					else
						s.setForeground(PLAYERS_COLOR);
	
					playersPan.add(s);
				}
	
			validate();
			repaint();

		} catch(NullPointerException e)  {}
	}

	@Override
	public void reset() {
		playersPan.removeAll();
		playPan.removeAll();
		boardPan.removeAll();

		proposedWord = false;
		letters = null;
		historiesBuffer = new HashMap<String, boolean[][]>();
		timer.setVisible(true);
		playPan.setVisible(true);
		wordField.setEditable(true);
	    playersPan.setVisible(false);

		outBanner();

		validate();
		repaint();
	}
	
	/**
	 * Popup A propos
	 */
	protected void about() {
		outInfo("Boggle", "Julien VAUBOURG / Igor MARTY\nCopyLeft (c) 2011");
	}

	protected void changeGameMode(Game.GAME_MODE gm) {
		// Si une partie est en cours, on demande la confirmation de son avortement
		if(game.getGameMode() != null) {
			if(inStop()) {
				game.interrupt();
				game_mode = gm;
			}
			
		// Sinon on demarre directement
		} else
			game.setGameMode(gm);
	}
	
	protected boolean inStop() {
		return JOptionPane.showConfirmDialog(
			this,
			"Arrêter la partie ?",
			"Changer de jeu",
			JOptionPane.YES_NO_OPTION
		) != JOptionPane.NO_OPTION;
	}

	@Override
	public void outNewRemotePlayer(String name) {}

	@Override
	public void outEE() {
		outInfo("Bravo", "Vous avez trouvé la réponse à la vie, l'univers et tout le reste... mais elle n'est pas dans le plateau.");
	}
}