/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.io.IOException;

public class Main {
	public static void main(String[] args) {
		
		// Controleur du jeu
		final Game g = new Game();

		// Ajout de la vue
		g.setIOFromArgs(args);
		g.getIO().outBanner();

		// Thread du jeu en cours (eventuellement en attente d'un choix de jeu)
		Runnable currentGame = new Runnable() {
			@Override
			public void run() {

				// Attend que le joueur n'indique le choix du jeu choisis
				g.getIO().inGameMode();

				try {
					// Jeu en ligne, version client
					if(g.getGameMode() == Game.GAME_MODE.ONLINE_CLIENT) {
						try {
							// Attend que le joueur client ait indique l'IP du serveur 
							g.getIO().inHostAddress();
	
							// Contacte le serveur (envoie une interruption si le serveur n'est pas joignable)
							g.setClientSocket();
							
							// Demande le nom du joueur client
							g.addPlayerClientLocal();
	
							// Envoi le nom du joueur au serveur
							g.getClientSocket().sendPlayer();
	
							// Attend que le serveur lance la partie, pour recuperer la liste des autres
							// joueurs ainsi que le plateau  (peut lancer une interruption si le serveur
							// devient injoignable)
							g.getIO().outBeginWaitServer();
							g.getClientSocket().receiptPlayers();
							g.getClientSocket().receiptBoard();
							g.getIO().outEndWaitServer();				

							// Lance en tache de fond le thread qui permettra de recevoir la synchro des scores
							// de tous les joueurs (toutes les secondes)
							NetworkSynchro.startClient(Network.getHost(), g);
	
							// Demarrage de la partie
							g.start();
							
							// Le joueur joue (peut lancer une interruption si le serveur devient injoignable)
							g.playing();

						// Le serveur est injoignable, remise a zero de la partie
						} catch (IOException e) {
							g.lostConnection();
						}
					
					// Jeu en mode serveur (en ligne, seul, ou a plusieurs sur la meme machine) 
					} else {
						try {
							
							// Charge le dictionnaire, si c'est la premiere partie en temps que serveur
							if(g.getLexicon().isEmpty()) {
								g.getIO().outBeginWaitLexicon();
								g.getLexicon().read("ODS5.txt.gz");
								g.getIO().outEndWaitLexicon();
							}

							// Attend que l'utilisateur ait indique un nom de joueur
							g.addPlayer();

							// Cree le plateau, puis le melange
							g.getBoard().setField("dice-definition.txt");
							g.mixup();
 
							switch(g.getGameMode()) {
							
								// Si c'est un jeu en ligne, un thread est lance en tache de fond
								// pour recevoir les joueurs distants, et communiquer ensuite avec eux
								// Un second thread s'occupe de la synchro permanente des scores des joueurs
								case ONLINE_HOST :
									g.startOnlineServer();
								break;

								// Si le jeu est local mais a plusieurs, on attend que l'utilisateur ait rentre
								// tous les noms joueurs qui desirent participer 
								case MULTI :
									g.addPlayers();
							}
	
							// Ajout du robot
							g.addBotPlayer();

							// Demarrage de la partie
							g.start();
							
							// Les joueurs sont invites a jouer tour a tour (si c'est en multi)
							g.playing();

							// L'utilisateur du serveur devra patienter tant que tout le monde n'a pas rendu
							// sa copie, et que le jeu n'a pas ete interrompu (sert pour le cas d'un serveur
							// en ligne)
							while (!g.isFinish() && !g.isInterrupted())
								try {
									Thread.sleep(200);
								} catch (InterruptedException e) {}

						// Des fichiers sont manquants : avortement du processus
						} catch (IOException e) {
							System.err.println("ODS5.txt.gz and/or dice-definition.txt not found !");
							System.exit(1);
						}
					}

				// La partie a ete interrompue (utilisation du menu principal au cours d'une partie,
				// perte de connexion, etc)
				} catch (InterruptedException e) {
					g.interrupt();
				}
			}
		};

		// Coeur du programme : des threads de jeu sont lances tant que l'utilisateur ne souhaite
		// pas profiter du soleil
		while(true)  {
			
			// Lancement d'un jeu
			Thread newGame = new Thread(currentGame);
			newGame.start();

			// Tant que le jeu n'a pas ete coupe
			while (newGame.isAlive()) {
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {}

				// Si une partie a ete lancee, le chrono est incremente toutes
				// les secondes, sauf en mode multi
				if(g.isStarted() && g.getGameMode() != Game.GAME_MODE.MULTI) {
					g.incrementTimer();
					g.getIO().outUpdateTimer();
				}

				// Si le chrono a atteint la valeur maximale, la partie est interrompue,
				// ce qui aura pour effet de terminer le thread et donc de sortir de la boucle
				if(g.getTimer() == Game.TIMER_MAX) {
					g.getIO().outElapsedTime();
					g.interrupt();
				}
			}

			/* Un jeu vient d'etre termine */

			// Si le jeu n'a pas ete interrompu brutalement, ou que c'est le chronometre qui en
			// a ete manifestement la cause (une partie vient de se terminer normalement)
			if(!g.isInterrupted() || g.getTimer() >= Game.TIMER_MAX) {
				
				// Si le joueur etait client d'une partie en ligne
				if(g.getGameMode() == Game.GAME_MODE.ONLINE_CLIENT) {
					try {
						// Reception de tous les mots trouves par les autres joueurs, depuis le serveur						
						g.getClientSocket().receiptWords();
						
					// Connexion perdue avec le serveur
					} catch (IOException e) {
						g.lostConnection();
					}
					
				// Le joueur etait serveur (en ligne ou non), le robot joue
				} else
					g.playBotPlayer();

				// Le joueur etait serveur dans une partie en ligne
				if(g.getGameMode() == Game.GAME_MODE.ONLINE_HOST) {
					
					// Interruption de l'attente de nouveaux mots de la part des joueurs distants,
					// pour tous threads serveurs
					NetworkServer.noMoreWords();

					// Tant que tous les joueurs distants n'ont pas recu les mots trouves par l'ensemble
					// des joueurs, on patiente (afin de ne pas couper les threads trop tot)
					while(!NetworkServer.isAllWordsSent())
						try {
							Thread.sleep(200);
						} catch (InterruptedException e) {}
				}

				// Si c'est une partie qui a ete completement finie, on affiche les resultats,
				// on remet a zero le jeu (en coupant notamment tous les threads reseaux) et on
				// attend que le joueur ne rechoisisse un mode de jeu
				if(g.isFinish()) {
					g.getIO().outResults();
					g.reset();
					g.getIO().inReplay();
				}

			// L'interruption a ete brutale : il n'y a rien a faire ni a conserver, on relance
			// simplement en nettoyant les traces de cette partie
			} else {
				g.reset();
				g.getIO().reset();
			}
		}
	}
}