/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.io.IOException;
import java.net.*;
import java.util.HashMap;

import org.ardverk.collection.PatriciaTrie;

public class NetworkClient extends Network {
	
	/**
	 * Cree une instance de communication avec le serveur, pour un client.
	 * @param game Controleur de jeu 
	 * @throws UnknownHostException
	 * @throws IOException
	 */
	public NetworkClient(Game game) throws UnknownHostException, IOException {
		this.game = game;
		
		// Etablissement de la connexion avec le serveur
		connection = new Socket(Network.getHost(), PORT_MAIN);
	}
	
	/**
	 * Decoupe une identite de joueur donnee en ses deux parties (IP et nom).
	 * @param identity Identite d'un joueur
	 * @return IP puis nom du joueur
	 */
	public String[] getRemoteIdentity(String identity) {
		return identity.split(PlayerRemote.IDENTITY_SEP, 2);
	}

	/**
	 * Envoi du nom du joueur client au serveur.
	 * @throws IOException
	 */
	public void sendPlayer() throws IOException {
		send(game.getPlayers().get(0).getName());
	}

	/**
	 * Proposition d'un mot a ajouter pour le joueur client, au serveur. Si la reponse de celui-ci
	 * est une WordException, elle est relancee. 
	 * @param word Mot a soumettre
	 * @param lp Joueur client
	 * @throws IOException
	 * @throws WordException
	 */
	public void sendWord(String word, PlayerClientLocal lp) throws IOException, WordException {
		send(word);

		Object e = receipt();

		if(e instanceof WordException)
			throw (WordException) e;
	}

	/**
	 * Reception de la liste des joueurs (identites) de la partie depuis le serveur.
	 * @throws IOException
	 */
	public void receiptPlayers() throws IOException {
		String[] identities = (String[]) receipt();

		// Pour chaque identite (IP + nom) de joueur recu, une instance est creee
		for(int i = 0; i < identities.length; i++) {
			String[] identity = getRemoteIdentity(identities[i]);

			// Si le joueur s'apparente au robot, on le defini en temps que tel
			if(identity[1].equals(PlayerBot.DEFAULT_NAME) && identity[0].equals(Player.DEFAULT_HOST))
				game.addPlayer(new PlayerBot(game));
			
			// Sinon on ajoute un joueur distant (pour le client)
			else
				game.addPlayer(new PlayerClientRemote(identity[1], game, identity[0]));

		}
	}

	/**
	 * Reception du plateau de la partie, depuis le serveur.
	 * @throws IOException
	 */
	public void receiptBoard() throws IOException {
		game.getBoard().setField((Dice[][]) receipt());
	}

	/**
	 * Reception de tous les mots trouves par les differents joueurs (listes identifiees
	 * avec les identites des joueurs).
	 * @doc http://stackoverflow.com/questions/262367/type-safety-unchecked-cast
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	public void receiptWords() throws IOException {
		HashMap<String, PatriciaTrie<String, Object>[]> words = (HashMap<String, PatriciaTrie<String, Object>[]>) receipt();

		// Attribution des listes des mots (tous, et originaux)
		for(Player p : game.getPlayers()) {
			
			// Si c'est un joueur distant on le robot
			if(p instanceof PlayerClientRemote)
				((PlayerClient) p).setWords(
					words.get(((PlayerClient) p).getRemoteIdentity())[0],
					words.get(((PlayerClient) p).getRemoteIdentity())[1]
				);
			else if(p instanceof PlayerBot) {
				((PlayerBot) p).setWords(words.get(p.getServerIdentity())[1]);
			} else
				((PlayerClient) p).setWords(
					words.get(((PlayerClient) p).getLocalIdentity())[0],
					words.get(((PlayerClient) p).getLocalIdentity())[1]
				);
			
			// Le joueur client a fini sa partie
			p.finish();
		}
	}
}