/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;

public class NetworkMultiSocketServer extends Network implements Runnable {
	static private NetworkMultiSocketServer multiMain = null;
	static private NetworkMultiSocketServer multiSynchro = null;
	private ArrayList<NetworkServer> servers = new ArrayList<NetworkServer>();
	private ArrayList<NetworkSynchro> synchros = new ArrayList<NetworkSynchro>();
	private ServerSocket socket;
	private volatile Thread thread;
	private int port;
	
	/**
	 * Cree un serveur multi socket permettant d'accepter les nouveaux joueurs et de leur
	 * attribuer un thread muni d'un tube (socket) directement relie a lui.
	 * @doc http://java.sun.com/developer/technicalArticles/ALT/sockets
	 * @doc http://edn.embarcadero.com/article/31995
	 * @param port Port a ecouter
	 * @param game Controleur du jeu
	 */
	public NetworkMultiSocketServer(int port, Game game) {
		this.port = port;
		this.game = game;
	}

	/**
	 * Permet de stopper le thread, en coupant au passage tous les threads crees par son biais.
	 * @doc http://download.oracle.com/javase/1.4.2/docs/guide/misc/threadPrimitiveDeprecation.html
	 */
	public void stop() {
		thread = null;

		// Si c'est le multisocket des threads de communication principaux, on les coupe
		if(port == PORT_MAIN)
			for(NetworkServer r : servers)
				r.stop();
		
		// Idem si ce sont ceux de synchro des scores
		else
			for(NetworkSynchro r : synchros)
				r.stop();

		// Fermeture du socket (avorte l'attente de nouveaux arrivants)
		try {
			socket.close();
		} catch (IOException e) {}
	}

	/**
	 * Retourne la liste des threads principaux crees
	 * @return Liste des threads principaux crees
	 */
	public ArrayList<NetworkServer> getServers() {
		return servers; 
	}
	
	/**
	 * Retourne le multisocket qui cree les threads principaux des joueurs distants. 
	 * @return Le multisocket qui cree les threads principaux des joueurs distants
	 */
	static public NetworkMultiSocketServer getMultiMain() {
		return multiMain; 
	}

	/**
	 * Lance en tache de fond le serveur qui cree les threads principaux des joueurs distants.
	 * @param game Controleur de jeu
	 */
	static void startMain(Game game) {
		multiMain = start(PORT_MAIN, game);
	}
	
	/**
	 * Stoppe le serveur qui cree les threads principaux des joueurs distants (et tue ses fils avec)
	 */
	static void stopMain() {
		if(multiMain != null)
			multiMain.stop();
	}

	/**
	 * Lance en tache de fond le serveur qui cree les threads de synchro des joueurs distants.
	 * @param game Controleur de jeu
	 */
	static void startSynchro(Game game) {
		multiSynchro = start(PORT_SYNCHRO, game);
	}
	
	/**
	 * Stoppe le serveur qui cree les threads de synchro des joueurs distants (et tue ses fils avec)
	 */
	static void stopSynchro() {
		if(multiSynchro != null)
			multiSynchro.stop();
	}

	/**
	 * Lance un multisocket.
	 * @param port Port a ecouter
	 * @param game Controleur de jeu
	 * @return Le multisocket lancé
	 */
	static NetworkMultiSocketServer start(int port, Game game) {
		Runnable runMultiSocket = new NetworkMultiSocketServer(port, game);
		new Thread(runMultiSocket).start();

		return (NetworkMultiSocketServer) runMultiSocket;
	}

	/**
	 * Corps du thread du multisocket.
	 */
	public void run() {
		// Permet de tuer le thread au moment souhaite (cf. le lien doc du constructeur)
		thread = Thread.currentThread();

		try {
			// Ecoute des requetes sur le port donne
			socket = new ServerSocket(port);
			
			try {
				// Tant que le thread n'est pas tue
				while (Thread.currentThread() == thread) {
					Thread.sleep(200);
	
					// Cas d'un multisocket d'ecoute pour les threads de synchro, ou que la
					// collecte des joueurs est active
					if(port == PORT_SYNCHRO || game.getAcceptRemotePlayers()) {
						
						// Attend et accepte toute nouvelle connexion
						Socket connection = socket.accept();
						Runnable runnable;

						// Cas d'un multisocket d'ecoute pour les threads principaux
						if(port == PORT_MAIN) {
							
							// Creation du thread dedie au joueur
							runnable = new NetworkServer(connection, game);
							servers.add((NetworkServer) runnable);
							
						// De synchro
						} else {
							
							// Creation du thread dedie au joueur
							runnable = new NetworkSynchro(connection, true, game);
							synchros.add((NetworkSynchro) runnable);
						}

						// Lancement du thread
						new Thread(runnable).start();
					}
				}

			// En cas de probleme de socket, le thread meurt simplement (interruption) 
			} catch(SocketException e) {}

		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}