/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.net.*;
import java.util.*;
import java.io.*;

import org.ardverk.collection.PatriciaTrie;

public class NetworkServer extends Network implements Runnable {
	private volatile Thread thread;
	private PlayerRemote player;
	private boolean isWordsSent;

	/**
	 * Cree une instance qui permettra de communiquer directement avec un joueur client
	 * pour lui transmettre toutes les informations de jeu, et recevoir ses informations.
	 * @param connection Socket de connexion avec le client
	 * @param game Controleur de jeu
	 */
	NetworkServer(Socket connection, Game game) {
		this.connection = connection;
		this.game = game;
	}

	/**
	 * Indique sur tous les joueurs distants ont recu les mots joueurs de la partie.
	 * @return Vrai si tous les joueurs distants ont recu les mots joueurs de la partie 
	 */
	static public boolean isAllWordsSent() {
		for(NetworkServer n : NetworkMultiSocketServer.getMultiMain().getServers())
			if(!n.isWordsSent())
				return false;
		
		return true;
	}
	
	/**
	 * Interrompt l'attente de nouveaux mots pour tous les threads principaux des joueurs clients. 
	 */
	static public void noMoreWords() {
		for(NetworkServer n : NetworkMultiSocketServer.getMultiMain().getServers())
			n.shutDownInput();
	}

	/**
	 * Tue le thread.
	 */
	public void stop() {
		isWordsSent = true;
		thread = null;
	}

	/**
	 * Rend le socket unilateral (permet de bloquer la reception de nouveaux mots, par exemple).
	 */
	private void shutDownInput() {
		try {
			connection.shutdownInput();
		} catch (IOException e) {}
	}
	
	/**
	 * Reception du nom de joueur du client concerne par le thread.
	 * @throws IOException
	 * @throws EOFException
	 */
	public void receiptPlayer() throws IOException, EOFException {
		String name = (String) receipt();

		// Creation du joueur client dans la partie
		player = new PlayerRemote(
			name, game,
			connection.getInetAddress().getHostAddress()
		);

		game.getIO().outNewRemotePlayer(player.getName());
		game.addPlayer(player);
	}

	/**
	 * Reception en continue des mots du joueur client, et renvoi eventuellement de la
	 * WordException qu'elle genere (sinon null).
	 * @throws IOException
	 * @throws EOFException
	 */
	public void receiptWords() throws IOException, EOFException {
		String word = (String) receipt();

		// Permet de tuer le thread
		while(!word.isEmpty() && !game.isInterrupted()) {

			try {
				// Si le mot ne renvoi pas d'exception, il est ajoute a la liste
				player.play(word);
				send(null);
				game.getIO().outUpdateScores();

			// Sinon, l'exception est relayee au joueur client a travers le reseau
			} catch(WordException e) {
				send(e);
			}

			word = (String) receipt();
		}

		send(null);
	}

	/**
	 * Envoi de la liste des joueurs de la partie au client.
	 * @throws IOException
	 */
	public void sendPlayers() throws IOException {
		ArrayList<Player> players = game.getPlayers();
		String[] names = new String[players.size() - 1];
		int i = 0;

		// Pour chaque joueur
		for(Player p : players)
			
			// Si c'est un autre joueur client, envoi de son identite avec son IP 
			if(p instanceof PlayerRemote) {
				if(p != player)
					names[i++] = ((PlayerRemote) p).getRemoteIdentity();

			// Sinon l'IP est remplacee par le nom par defaut du serveur
			} else if(p instanceof PlayerHuman || p instanceof PlayerBot)
				names[i++] = p.getServerIdentity();

		send(names);
	}

	/**
	 * Envoi de la liste des mots trouves par les differents joueurs (y compris ceux du joueur client).
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	public void sendWords() throws IOException {
		HashMap<String, PatriciaTrie<String, Object>[]> words = new HashMap<String, PatriciaTrie<String, Object>[]>();

		// Pour chaque joueur
		for(Player p : game.getPlayers()) {
			PatriciaTrie[] p_words = {
				p.getWords(),
				p.getOriginalWords()
			};

			// Association des listes de mots aux identites des joueurs
			if(p instanceof PlayerRemote)
				if(p == player)
					words.put(((PlayerRemote) p).getLocalIdentity(), p_words);
				else
					words.put(((PlayerRemote) p).getRemoteIdentity(), p_words);
			else
				words.put(p.getServerIdentity(), p_words);
		}

		// Envoi des associations
		send(words);
	}

	/**
	 * Indique si le thread a eu le temps d'envoyer la liste des mots au client.
	 * @return Vrai si le thread a eu le temps d'envoyer la liste des mots au client
	 */
	public boolean isWordsSent() {
		return isWordsSent;
	}

	/**
	 * Envoi du plateau au joueur client.
	 * @throws IOException
	 */
	public void sendBoard() throws IOException {
		send(game.getBoard().getField());
	}
	
	/**
	 * Corps du thread.
	 */
	public void run() {
		thread = Thread.currentThread();
		
		try {
			// Permet de tuer le thread, et de traiter les taches en continue
			while (Thread.currentThread() == thread) {
				Thread.sleep(200);

				// Si on est encore en plein dans la collecte des nouveaux joueurs,
				// on recoit le nom du joueur pour le creer dans le jeu
				if(game.getAcceptRemotePlayers()) {
					try {
						receiptPlayer();

						// Attente du depart du jeu pour continuer les actions
						while (!game.isStarted())
							Thread.sleep(200);

						// Envoi de la listes des joueurs au joueur client
						sendPlayers();

						// Envoi du plateau
						if(Thread.currentThread() == thread)
							sendBoard();

					} catch(EOFException e) {
						stop();
					}

				// Si le jeu est fini, envoi des mots et fin du thread
				} else if(game.isFinish()) {
					sendWords();
					isWordsSent = true;
					break;

				// Si le jeu n'est pas fini et que le joueur client n'a pas fini non plus
				} else if(!player.isFinish()) {
					try {
						
						// Reception des mots proposes par le joueur distant
						receiptWords();
					} catch(EOFException e) {}

					// Le joueur a indique qu'il a fini de jouer
					player.finish();
				}
			}

		} catch (IOException e) {
			e.printStackTrace();

		} catch (InterruptedException e) {
		} finally {
			disconnect();
		}
	}
}