/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.io.EOFException;
import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.HashMap;

public class NetworkSynchro extends Network implements Runnable {
	private boolean isServer;
	private volatile Thread thread;
	private HashMap<String, Integer> scores;
	static private NetworkSynchro client = null;

	/**
	 * Cree une nouvelle instance de pour un thread de synchronisation des scores
	 * (client ou serveur).
	 * @param connection Connexion avec le client/serveur
	 * @param isServer Vrai si c'est le serveur qui cree cette instance
	 * @param game Controleur du jeu
	 */
	public NetworkSynchro(Socket connection, boolean isServer, Game game) {
		this.connection = connection;
		this.isServer = isServer;
		this.game = game;		
	}

	/**
	 * Tue le thread.
	 */
	public void stop() {
		thread = null;
	}

	/**
	 * Lance un client de synchronisation des scores en tache de fond.
	 * @param host IP du serveur hote
	 * @param game Controleur de jeu
	 * @throws UnknownHostException
	 * @throws IOException
	 */
	static void startClient(String host, Game game) throws UnknownHostException, IOException {
		Socket connection = new Socket(host, PORT_SYNCHRO);
		Runnable runSynchro = new NetworkSynchro(connection, false, game);
		new Thread(runSynchro).start();
	}
	
	/**
	 * Tue la version client.
	 */
	static void stopClient() {
		if(client != null)
			client.stop();
	}

	/**
	 * Envoi de la liste des scores au client concerne par le thread.
	 * @throws IOException
	 */
	private void sendScores() throws IOException {
		scores = new HashMap<String, Integer>();

		// Pour chaque joueur, ajout de son score dans la liste (referencee par identite)
		for(Player p : game.getPlayers()) {
			if(p instanceof PlayerRemote) {
				
				// Si le joueur correspond à celui qui va recevoir la liste, on lui sert une identite
				// qui lui permettra de se reconnaitre (impossible parfois pour lui de connaitre sa propre
				// adresse IP)
				if(((PlayerRemote) p).equals(connection.getInetAddress()))
					scores.put(((PlayerRemote) p).getLocalIdentity(), p.getScore());
				
				// Sinon, on envoi la version distante
				else
					scores.put(((PlayerRemote) p).getRemoteIdentity(), p.getScore());

			// Ou encore la version avec le tag serveur
			} else if(p instanceof PlayerHuman)
				scores.put(p.getServerIdentity(), p.getScore());
		}

		send(scores);
	}

	/**
	 * Reception de la liste des scores des joueurs par le serveur.
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	private void receiptScores() throws IOException {
		try {
			scores = (HashMap<String, Integer>) receipt();

			// Si le jeu n'est pas fini, on met a jour avec ce qu'on a recu
			if(!game.isFinish()) {
				for(Player p : game.getPlayers()) {
					if(!(p instanceof PlayerBot)) {
						if(p instanceof PlayerClientRemote)
							((PlayerClient) p).setScore(scores.get(((PlayerClient) p).getRemoteIdentity()));
						else
							((PlayerClient) p).setScore(scores.get(((PlayerClient) p).getLocalIdentity()));
					}
				}
			
				game.getIO().outUpdateScores();
			}

		// Biais de synchro
		} catch(NullPointerException e) { 
		} catch(EOFException e) {}		
	}

	/**
	 * Corps du thread.
	 */
	@Override
	public void run() {
		thread = Thread.currentThread();
		
		try {
			// Permet de tuer le thread
			while (Thread.currentThread() == thread) {
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {}

				if(isServer)
					sendScores();
				else
					receiptScores();
			}

		} catch (IOException e) {
			e.printStackTrace();

		} finally {
			disconnect();
		}
	}
}