/*
 * Boggle - TOP PROJET ESIAL 2011
 * Copyright (C) 2011 Julien VAUBOURG / Igor MARTY
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package boggle;

import java.io.IOException;
import org.ardverk.collection.*;

public abstract class Player implements Comparable<Player> {
	public final static String DEFAULT_HOST = "Server";
	protected String name;
	protected PatriciaTrie<String, Object> allWordsFound;
	protected PatriciaTrie<String, Object> originalWordsFound;
	protected Game game;
	protected boolean finish;

	/**
	 * Cree un joueur (depuis un heritage).  
	 * @param name Nom du joueur
	 * @param game Controleur du jeu
	 */
	public Player(String name, Game game) {
		this.name = name;
		this.game = game;
		allWordsFound = new PatriciaTrie<String, Object>(StringKeyAnalyzer.INSTANCE);
		originalWordsFound = new PatriciaTrie<String, Object>(StringKeyAnalyzer.INSTANCE);
	}

	/**
	 * Ajouter un mot a la liste des mots originaux du joueur.
	 * @param word Mot a ajouter
	 */
	public void add2OriginalWords(String word) {
		originalWordsFound.put(word, null);
	}
	
	/**
	 * Ajouter un mot a la liste exaustive des mots trouves par le joueur.
	 * @param word Mot a ajouter
	 */
	public void add2Words(String word) {
		allWordsFound.put(word, null);
	}

	/**
	 * Permet de trier les joueurs par score.
	 */
	public int compareTo(Player p) {
		if(getScore() == p.getScore())
			return 0;

		return getScore() > p.getScore() ? 1 : -1;
	}

	/**
	 * Retourne le nom du joueur.
	 * @return Nom du joueur
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Retourne le score du joueur
	 * @return Score du joueur
	 */
	public int getScore() {
		int score = 0;

		for(String word : originalWordsFound.keySet())
			score += game.wordScore(word);

		return score;
	}

	/**
	 * Ajoute un mot a la liste de tous les mots du joueurs, et eventuellement a celle
	 * des mots originaux si personne d'autre ne l'a trouve avant. Si c'est le cas,
	 * sort aussi le mot en question en la liste des mots originaux du joueur concerne. 
	 * @param word Mot a ajouter
	 */
	public void addWord(String word) {
		boolean alreadyFound = false;

		// Pour chaque joueur, si le mot a deja ete trouve, il ne devient plus
		// original pour lui
		for(Player p : game.getPlayers())
			if(p instanceof PlayerHuman && p.getWords().containsKey(word)) {
				alreadyFound = true;
				p.markAsNotOriginal(word);
			}
		
		// Ajout du mot a la liste exaustive
		add2Words(word);

		// Ajout a la liste des originaux, si personne ne l'a trouve avant
		if(!alreadyFound)
			add2OriginalWords(word);
	}

	/**
	 * Sort un mot de la liste des mots originaux du joueur.
	 * @param word Mot a sortir
	 */
	public void markAsNotOriginal(String word) {
		originalWordsFound.remove(word);
	}

	/**
	 * Retourne la liste des mots originaux trouves par le joueur.
	 * @return La liste des mots originaux trouves par le joueur
	 */
	public PatriciaTrie<String, Object> getOriginalWords() {
		return originalWordsFound;
	}

	/**
	 * Retourne la liste exaustive des mots trouves par le joueur.
	 * @return La liste exaustive des mots trouves par le joueur
	 */
	public PatriciaTrie<String, Object> getWords() {
		return allWordsFound;
	}
	
	/**
	 * Le joueur a fini de jouer.
	 */
	public void finish() {
		finish = true;
	}
	
	/**
	 * Indique si le joueur a fini de jouer.
	 * @return Vrai si le joueur a fini de jouer
	 */
	public boolean isFinish() {
		return finish;
	}
	
	/**
	 * Affichage du nom du joueur.
	 */
	public String toString() {
		return name;
	}
	
	/**
	 * Retourne l'identite du joueur : sans surcharge, renvoi le DEFAULT_HOST en tant qu'IP
	 * (identification des joueurs du serveur).
	 * @return Identite par defaut du joueur
	 */
	public String getServerIdentity() {
		return DEFAULT_HOST + PlayerRemote.IDENTITY_SEP + name;
	}

	/**
	 * Verifie si le mot peut bien etre ajoute a la liste des mots trouves (lisible sur le plateau,
	 * taille minimum, present dans le dictionnaire). Renvoi une WordException si ca n'est pas le cas.
	 * @param word Mot a verifier
	 * @throws WordException
	 * @throws IOException
	 */
	public abstract void checkWord(String word) throws WordException, IOException;
	
	/**
	 * Permet au joueur de jouer.
	 * @throws IOException
	 * @throws InterruptedException
	 */
	public abstract void play() throws IOException, InterruptedException;
}